import torch
import torch.nn as nn
import torch.optim as optim

import torch.nn.functional as F
import torchaudio
import tqdm

import argparse

import math

import os

from torch.utils.data import random_split
from torch.utils.data import DataLoader

import AudioUtils
from data import AudioDataset
from generator import SISUGenerator
from discriminator import SISUDiscriminator

import torchaudio.transforms as T

# Init script argument parser
parser = argparse.ArgumentParser(description="Training script")
parser.add_argument("--generator", type=str, default=None,
                    help="Path to the generator model file")
parser.add_argument("--discriminator", type=str, default=None,
                    help="Path to the discriminator model file")
parser.add_argument("--device", type=str, default="cpu",  help="Select device")
parser.add_argument("--epoch", type=int, default=0, help="Current epoch for model versioning")
parser.add_argument("--verbose", action="store_true",  help="Increase output verbosity")

args = parser.parse_args()

device = torch.device(args.device if torch.cuda.is_available() else "cpu")
print(f"Using device: {device}")

mfcc_transform = T.MFCC(
    sample_rate=44100,  # Adjust to your sample rate
    n_mfcc=20,
    melkwargs={'n_fft': 2048, 'hop_length': 512} # adjust n_fft and hop_length to your needs.
).to(device)

def gpu_mfcc_loss(y_true, y_pred):
    mfccs_true = mfcc_transform(y_true)
    mfccs_pred = mfcc_transform(y_pred)
    min_len = min(mfccs_true.shape[2], mfccs_pred.shape[2])
    mfccs_true = mfccs_true[:, :, :min_len]
    mfccs_pred = mfccs_pred[:, :, :min_len]
    return torch.mean((mfccs_true - mfccs_pred)**2)

def discriminator_train(high_quality, low_quality, real_labels, fake_labels):
    optimizer_d.zero_grad()

    # Forward pass for real samples
    discriminator_decision_from_real = discriminator(high_quality[0])
    d_loss_real = criterion_d(discriminator_decision_from_real, real_labels)

    # Forward pass for fake samples (from generator output)
    generator_output = generator(low_quality[0])
    discriminator_decision_from_fake = discriminator(generator_output.detach())
    d_loss_fake = criterion_d(discriminator_decision_from_fake, fake_labels)

    # Combine real and fake losses
    d_loss = (d_loss_real + d_loss_fake) / 2.0

    # Backward pass and optimization
    d_loss.backward()
    nn.utils.clip_grad_norm_(discriminator.parameters(), max_norm=1.0)  # Gradient Clipping
    optimizer_d.step()

    return d_loss

def generator_train(low_quality, high_quality, real_labels):
    optimizer_g.zero_grad()

    # Forward pass for fake samples (from generator output)
    generator_output = generator(low_quality[0])

    mfcc_l = gpu_mfcc_loss(high_quality[0], generator_output)

    discriminator_decision = discriminator(generator_output)
    adversarial_loss = criterion_g(discriminator_decision, real_labels)

    combined_loss = adversarial_loss + 0.5 * mfcc_l

    combined_loss.backward()
    optimizer_g.step()

    return (generator_output, combined_loss, adversarial_loss, mfcc_l)

debug = args.verbose

# Initialize dataset and dataloader
dataset_dir = './dataset/good'
dataset = AudioDataset(dataset_dir, device)

# ========= SINGLE =========

train_data_loader = DataLoader(dataset, batch_size=128, shuffle=True)

# Initialize models and move them to device
generator = SISUGenerator()
discriminator = SISUDiscriminator()

epoch: int = args.epoch

generator = generator.to(device)
discriminator = discriminator.to(device)

if args.generator is not None:
    generator.load_state_dict(torch.load(args.generator, map_location=device, weights_only=True))
if args.discriminator is not None:
    discriminator.load_state_dict(torch.load(args.discriminator, map_location=device, weights_only=True))

# Loss
criterion_g = nn.MSELoss()
criterion_d = nn.BCELoss()

# Optimizers
optimizer_g = optim.Adam(generator.parameters(), lr=0.0001, betas=(0.5, 0.999))
optimizer_d = optim.Adam(discriminator.parameters(), lr=0.0001, betas=(0.5, 0.999))

# Scheduler
scheduler_g = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer_g, mode='min', factor=0.5, patience=5)
scheduler_d = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer_d, mode='min', factor=0.5, patience=5)

models_dir = "models"
os.makedirs(models_dir, exist_ok=True)

def start_training():
    generator_epochs = 5000
    for generator_epoch in range(generator_epochs):
        low_quality_audio = (torch.empty((1)), 1)
        high_quality_audio = (torch.empty((1)), 1)
        ai_enhanced_audio = (torch.empty((1)), 1)

        times_correct = 0

        # ========= TRAINING =========
        for high_quality_clip, low_quality_clip in tqdm.tqdm(train_data_loader, desc=f"Training epoch {generator_epoch+1}/{generator_epochs}, Current epoch {epoch+1}"):
        # for high_quality_clip, low_quality_clip in train_data_loader:
            high_quality_sample = (high_quality_clip[0], high_quality_clip[1])
            low_quality_sample = (low_quality_clip[0], low_quality_clip[1])

            # ========= LABELS =========
            batch_size = high_quality_clip[0].size(0)
            real_labels = torch.ones(batch_size, 1).to(device)
            fake_labels = torch.zeros(batch_size, 1).to(device)

            # ========= DISCRIMINATOR =========
            discriminator.train()
            d_loss = discriminator_train(high_quality_sample, low_quality_sample, real_labels, fake_labels)

            # ========= GENERATOR =========
            generator.train()
            generator_output, combined_loss, adversarial_loss, mfcc_l = generator_train(low_quality_sample, high_quality_sample, real_labels)

            if debug:
                print(d_loss, combined_loss, adversarial_loss, mfcc_l)
            scheduler_d.step(d_loss)
            scheduler_g.step(combined_loss)

            # ========= SAVE LATEST AUDIO =========
            high_quality_audio = (high_quality_clip[0][0], high_quality_clip[1][0])
            low_quality_audio = (low_quality_clip[0][0], low_quality_clip[1][0])
            ai_enhanced_audio = (generator_output[0], high_quality_clip[1][0])

        new_epoch = generator_epoch+epoch

        if generator_epoch % 10 == 0:
            print(f"Saved epoch {new_epoch}!")
            torchaudio.save(f"./output/epoch-{new_epoch}-audio-crap.wav", low_quality_audio[0].cpu(), high_quality_audio[1]) # <-- Because audio clip was resampled in data.py from original to crap and to original again.
            torchaudio.save(f"./output/epoch-{new_epoch}-audio-ai.wav", ai_enhanced_audio[0].cpu(), ai_enhanced_audio[1])
            torchaudio.save(f"./output/epoch-{new_epoch}-audio-orig.wav", high_quality_audio[0].cpu(), high_quality_audio[1])

        if debug:
            print(generator.state_dict().keys())
            print(discriminator.state_dict().keys())
        torch.save(discriminator.state_dict(), f"{models_dir}/discriminator_epoch_{new_epoch}.pt")
        torch.save(generator.state_dict(), f"{models_dir}/generator_epoch_{new_epoch}.pt")

    torch.save(discriminator, "models/epoch-5000-discriminator.pt")
    torch.save(generator, "models/epoch-5000-generator.pt")
    print("Training complete!")

start_training()
